/**<pre>
 * Title: PhoneApp
 * Copyright: Copyright (c) 2003
 * Company:   University of Oregon Computer Science Dept.
 *
 * @author Jason Prideaux
 * @version 1.0
 * </pre>
 */

package app;

import app.util.*;

import cat.app.*;
import cat.context.*;
import cat.sig.*;
import cat.sig.event.*;
import cat.hist.*;
import cat.sensor.core.*;
import cat.sensor.util.*;
import cat.sensor.event.*;
import cat.sensor.soft.*;
import cat.util.*;
import cat.transport.*;
import transport.protocol.*;

import java.util.*;
import java.io.*;
import java.net.*;

/* ===================================================================== */
/** This class is a simple application for testing the wearable context
 *  system.  It should create a ContextManager that it will use for
 *  getting local and remote sensor data.  This class will also create
 *  the SIGs that it wants to be a part of.
 *
 *  If this application wants to use CAT, then it should extend ContextApp.
 *
 *  In this app we form a sig with everyone that we encounter, then we display
 *  whether the people in the sig are available to be called.
 *
 */
public class PhoneApp extends ContextApp{


	//The most current value of the OKTOCALL sensor.
	private HashMap table = new HashMap();
	private HashMap table2 = new HashMap();

	private PhoneGUI gui;

    /* ===================================================================== */
    /** This constructor will start the phone application.  It should call the
     *  setup method.
     *
     */
	public PhoneApp(String id, int port){

		super(id,port);

		init();

		gui = new PhoneGUI(this, id);

	} //Constructor



    /* ===================================================================== */
    /** Here is where we create any services/soft sensors that this application
     *  will need.
     *
     *  We can also start some sigs in here if we wanted to.
     *
     */
	public void init(){

		//This app requires special soft sensors/services.
		//so make them here.  This how to create our own soft
		//sensor service dynamically.
		Sensor sensor = getContextManager().getSensorManager().getSensor("SOUND");
		CallableSensor oktocall = new CallableSensor("OKTOCALL");
		oktocall.setSensorType( SensorOntology.OKTOCALL );
		sensor.addSensorListener(oktocall);
		//oktocall.addSensorListener(getContextManager().getHistoryManager());
		//getContextManager().getSensorManager().addSensor("OKTOCALL", oktocall);
		getContextManager().addSensor(oktocall);



	} //method: appSetUp



    /* ===================================================================== */
    /** This method checks if it is okay to make a phone call, if so then it
     *  will make the phone call.
     *  This method could be called by a button in the GUI for this app.
     *
     */
	public void tryPhoneCall( String id ){

		Agent B = (Agent)(table2.get(id));

		boolean av = checkIfAgentAvailable( id );
		if( av ){
			System.out.println("   The agent is ok to call.");
			makePhoneCall( B );
		}else{
			gui.setStat("The agent is too busy right now.");
			System.out.println("   The agent is too busy right now.");
		}

	} //method: tryPhoneCall



    /* ===================================================================== */
    /** This method will look through the current remote data we have collected.
     *  Either, we get the data from the ContextManager or we collect it ourself
     *  from the handleContextEvent method.
     *
     */
	public boolean checkIfAgentAvailable( String id ){

		//look at current sensor data
		return ((Boolean)(table.get(id))).booleanValue();

	} //method: checkIfAgentAvailable



    /* ===================================================================== */
    /** This method is used to place a phone call to the given agent.
     *  A phone call is simulated with a Socket.
     *
     */
	public void makePhoneCall( Agent B ){

		try{

			gui.setStat("Making the phone call to "+ B.getId() +".");

			UnicastMessage msg = new UnicastMessage(""+SensorOntology.P2PDATA, B.getPort(), iSIM.PORT );
			msg.setTo(B.getPort());
			msg.setSender(new Agent(iSIM.ID, iSIM.PORT ));
			msg.setContent( "phonecall" );

			//use transport to send it.
			//we don't care how it gets sent.
			new Transport().sendMessage(msg);


		}catch(Exception e){
			System.out.println("      No dial tone.");
		}

	}


    /* ===================================================================== */
    /** We are asked by this method if this application would like to join
     *  a sig.
     *
     *  @param  name   The name of the SIG.
     *  @param  agent   The agent who wants us to join the SIG.
     *  @param
     *  @param
     *
     */
	public boolean joinSig(String name, Vector agents, Vector wanted_sensors, Vector sharing_sensors){

		//return true if we want to join the sig.
		//false otherwise.

		return true;
	}

	public void joinedSig(String name, Agent agent){}


    /* ===================================================================== */
    /** This method is required by ContextApp.  This is where we are supposed
     *  to process sensor data that comes from local and remote sensors.
     *
     *  For this app we only care about the remote OKTOCALL sensor, so that
     *  we can know if it is okay to call some other agent.
     *
     */
	public void run(){

		while(true){
			try{Thread.sleep(200);}catch(Exception e){}


		  if(!getSensorQueue().isEmpty()){
			Object event = getSensorQueue().remove(0);
			SensorEvent se = (SensorEvent) event;
			if (se instanceof SensorDataEvent){
			  SensorDataEvent dataE = (SensorDataEvent) se;

			  String id = (String)(dataE.getSource());
			  int type = dataE.getType();
			  Value value = dataE.getValue();

			  //check who the message is from, and record the value.
			  //System.out.println("Receiving Sensor data from: "+id );

			  //if( type == SensorOntology.OKTOCALL && id.indexOf("remote_phone") > -1 ){
			  if( type == SensorOntology.OKTOCALL && dataE.isRemote() ){

				  System.out.println("PhoneApp: Getting remote data from sensor OKTOCALL.");
				  System.out.println("   from Agent "+dataE.getAgent().getId() +"  at "+ dataE.getAgent().getPort() );
				  //available = ((Boolean)(value.getValue())).booleanValue();
				  table.put(dataE.getAgent().getId(), (Boolean)(value.getValue()) );
				  table2.put(dataE.getAgent().getId(), dataE.getAgent() );
				  gui.update(table);

			  }else if( type == SensorOntology.P2PDATA && dataE.isRemote() ){


				  if( ((String)(value.getValue())).equals("phonecall") ){

					  gui.setStat("Receiving phone call from "+ dataE.getAgent().getId()+".");

					  try{
							try{Thread.sleep(4000);}catch(Exception e){}

							//either take the call or deicide if we can call back.
							System.out.println("   Checking to see if we can call back later.");

							//create the time when we want to call back.
							Calendar calendar = new GregorianCalendar();
							calendar.add(Calendar.MINUTE, 30);       //in 30 minutes.

							Object obj = getContextManager().getHistoryManager().query( "OKTOCALL", SensorOntology.OKTOCALL, calendar );

							try{
								if( (((Boolean)obj).booleanValue()) ){
									System.out.println("      You can call back in 30 minutes.");
									gui.setStat("You can call back in 30 minutes.");

									UnicastMessage msg = new UnicastMessage(""+SensorOntology.P2PDATA, dataE.getAgent().getPort(), iSIM.PORT );
									msg.setTo(dataE.getAgent().getPort());
									msg.setSender(new Agent(iSIM.ID, iSIM.PORT ));
									msg.setContent( "noanswer" );
									new Transport().sendMessage(msg);

								}else if( !(((Boolean)obj).booleanValue()) ){
									System.out.println("      You should take the call now.");
									gui.setStat("You will take the call now.");

									UnicastMessage msg = new UnicastMessage(""+SensorOntology.P2PDATA, dataE.getAgent().getPort(), iSIM.PORT );
									msg.setTo(dataE.getAgent().getPort());
									msg.setSender(new Agent(iSIM.ID, iSIM.PORT ));
									msg.setContent( "takecall" );
									new Transport().sendMessage(msg);
								}
							}catch(Exception e){
								System.out.println("      You should take the call now..");
								gui.setStat("You will take the call now.");
									UnicastMessage msg = new UnicastMessage(""+SensorOntology.P2PDATA, dataE.getAgent().getPort(), iSIM.PORT );
									msg.setTo(dataE.getAgent().getPort());
									msg.setSender(new Agent(iSIM.ID, iSIM.PORT ));
									msg.setContent( "takecall" );
									new Transport().sendMessage(msg);
								e.printStackTrace();
							}

						}catch(Exception ee){System.out.println("TakeCall: "+ee);}

				  }else if( ((String)(value.getValue())).equals("willtake") ){
					  gui.setStat("Your phone call was accepted by "+dataE.getAgent().getId()+".");

				  }else if( ((String)(value.getValue())).equals("noanswer") ){
					  gui.setStat(""+dataE.getAgent().getId() +" will call you back later.");
				  }


			  }else  if( type == SensorOntology.DISCOVERY ){
				  System.out.println("PhoneApp: Getting current list of agents near us:");
				  Agent[] agents = (Agent[])(value.getValue());
				  if( agents.length == 0 ){
						System.out.println("   empty");
				  }else{
				  		for( int i=0; i<agents.length; i++){
					  		System.out.println("   "+agents[i].getId() +" "+ agents[i].getPort() );
						}
				  }
			  }else  if( type == SensorOntology.ENCOUNTER ){
				  System.out.println("PhoneApp: Getting an encounter event:");
				  Agent agent = (Agent)(value.getValue());

				  if( getContextManager().getSigManager().findSig("phone") != null ){
					  Vector want = new Vector();
					  want.add( new Integer(SensorOntology.OKTOCALL) );
					  getContextManager().getSigManager().addAgentToSig("phone", agent, want);
				  }else{
					  formSig(agent);
				  }
				  System.out.println("   encountered agent " +agent);

			  }

			}
		  }
		}
	}


	public void formSig(Agent a){

		if( iSIM.PORT == 6000 ){

			//pick a sensor that this app will share.
			Sensor toshare = null;

			//Need to find the OKTOCALL sensor that we'll share
			Sensor[] sensors = getContextManager().getSensorManager().getAllSensors();
			for(int i=0; i<sensors.length; i++ ){

				//want to share just this sensor
				if( sensors[i].getSensorType() == SensorOntology.OKTOCALL )
					toshare = sensors[i];
			}


			//create a sig group and share one sensor.
			SigManager sigman = getContextManager().getSigManager();

			Vector sharing = new Vector();
			sharing.add( toshare );
			Vector want = new Vector();
			want.add( new Integer(SensorOntology.OKTOCALL) );

			//System.out.println("Alice is creating a SIG with Bob in it.");
			//System.out.println("   Sending a SigRequest.");
			sigman.createSig("phone", a, want, sharing );
		}

	}


	public static void main(String[] args){

		String id = CommandLine.getArg(args, "a", "bob");
		String port = CommandLine.getArg(args, "p", "6001");

		//run PhoneApp with bob.
		new PhoneApp(id, Integer.parseInt(port));

	} //method: main


} //class: PhoneApp
